# distutils: libraries = symmetrica
"""
Symmetrica library
"""

include "symmetrica.pxi"

include "kostka.pxi"

include "sab.pxi"

include "sc.pxi"

include "sb.pxi"

include "part.pxi"

include "schur.pxi"

include "plet.pxi"
