r"""
Named matroids

A module containing several interesting matroids that have names in the
literature. A list of all matroids in this database is available via tab
completion. Type ``matroids.named_matroids.`` (include the last dot) and hit
the ``tab`` key to see which matroids are available.
"""
from sage.matroids.catalog import R10, R12, Fano, NonFano, P6, R6, Q6, P7, Vamos, NonVamos, Pappus, NonPappus, R8, J, T8, S8, TicTacToe, Q10, N1, N2, BetsyRoss, Block_9_4, Block_10_5, ExtendedBinaryGolayCode, ExtendedTernaryGolayCode, AG23minus, NotP8, P8, P8pp, D16, Terrahawk, R9A, R9B, T12, P9, O7, AG32prime, F8, Q8, L8, K33dual, TernaryDowling3
