from sage.misc.lazy_import import lazy_import as _lazy_import_

_lazy_import_("sage.lfunctions.dokchitser", "Dokchitser")
_lazy_import_("sage.lfunctions.sympow", "sympow")
_lazy_import_("sage.lfunctions.zero_sums", "LFunctionZeroSum")
_lazy_import_("sage.lfunctions.lcalc", "lcalc")
