"""
Combinatorial Hopf Algebras
"""
from __future__ import absolute_import
from sage.misc.lazy_import import lazy_import

lazy_import('sage.combinat.chas.fsym', ['FreeSymmetricFunctions'])
lazy_import('sage.combinat.chas.wqsym', ['WordQuasiSymmetricFunctions'])
