r"""
Combinatorics on matrices
=========================

- :ref:`sage.combinat.matrices.dancing_links`
- :ref:`sage.combinat.matrices.dlxcpp`
- :ref:`sage.combinat.matrices.hadamard_matrix`
- :ref:`sage.combinat.matrices.latin`
"""
