;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((nil
  ;; Use space instead of tabs for indentation
  (indent-tabs-mode . nil))
 (makefile-mode
  ;; But use tabs in Makefiles
  (indent-tabs-mode . t)))
